// SPDX-License-Identifier: GPL-2.0
/*
 * Memory Controller L3 cache profile driver for Mellanox BlueField SoC
 *
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/acpi.h>
#include <linux/arm-smccc.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>

#include <uapi/linux/psci.h>

#include "l3cache_regs.h"

#define DRIVER_NAME		"mlx-l3cache"
#define DRIVER_VERSION		"0.2"
#define DRIVER_DESCRIPTION	"Mellanox BlueField SoC L3 Cache Profile Driver"

/* SMC return codes */
#define SMCCC_ACCESS_VIOLATION (-4)

/* SMC function identifiers */
#define MLNX_WRITE_REG_32		(0x82000009)
#define MLNX_SIP_SVC_UID		(0x8200ff01)
#define MLNX_SIP_SVC_VERSION		(0x8200ff03)

#define MLNX_L3C_SVC_REQ_MAJOR 0
#define MLNX_L3C_SVC_MIN_MINOR 3

#define L3_PER_MSS	2
#define L3_PROFILE_NUM	(L3C_PROF_RD_MISS__LENGTH / L3C_PROF_RD_MISS__STRIDE)

/* The offset of the given L3 register with regards to it's index. */
#define L3_REG_OFFSET(reg, idx)		\
	((L3C_##reg##__FIRST_WORD + L3C_##reg##__STRIDE * (idx)))

enum l3_profile_type {
	LRU_PROFILE = 0,	/* 0 is the default behavior. */
	NVME_PROFILE,
	L3_PROFILE_TYPE_NUM,
};

struct l3_profile {
	const char *name;	/* Name of the profile */
	L3C_PROF_RD_MISS_t prm;	/* Register values for this profile. */
	L3C_PROF_WR_MISS_t pwm;
	L3C_PROF_RD_HIT0_t prh0;
	L3C_PROF_WR_HIT0_t pwh0;
	L3C_PROF_RD_HIT1_t prh1;
	L3C_PROF_WR_HIT1_t pwh1;
};

static const struct l3_profile l3_profiles[L3_PROFILE_TYPE_NUM] = {
	[LRU_PROFILE] = {
		.name = "Strict_LRU",
		.prm = { /* L3C_PROF_RD_MISS_t */
			.low_order		= 0xb,
			.high_order		= 0x0,
			.alloc_state		= 0x0,
			.low_state_blk_alloc	= 0x0,
			.high_state_blk_alloc	= 0x0,
			.prob			= 0x0,
		},
		.pwm = { /* L3C_PROF_WR_MISS_t */
			.low_order		= 0xb,
			.high_order		= 0x0,
			.alloc_state		= 0x0,
			.low_state_blk_alloc	= 0x0,
			.high_state_blk_alloc	= 0x0,
			.prob			= 0x0,
		},
		.prh0 = { /* L3C_PROF_RD_HIT0_t */
			.low_delta		= 0xb,
			.high_delta		= 0x0,
			.new_state		= 0x0,
			.prob			= 0x0,
		},
		.pwh0 = { /* L3C_PROF_WR_HIT0_t */
			.low_delta		= 0xb,
			.high_delta		= 0x0,
			.new_state		= 0x0,
			.prob			= 0x0,
		},
		.prh1 = { /* L3C_PROF_RD_HIT1_t */
			.low_delta		= 0xb,
			.high_delta		= 0x0,
			.new_state		= 0x0,
			.prob			= 0x0,
		},
		.pwh1 = { /* L3C_PROF_WR_HIT1_t */
			.low_delta		= 0xb,
			.high_delta		= 0x0,
			.new_state		= 0x0,
			.prob			= 0x0,
		},
	},
	[NVME_PROFILE] =  {
		.name = "NVMeOF_suitable",
		.prm = { /* L3C_PROF_RD_MISS_t */
			.low_order		= ~0,
			.high_order		= ~0,
			.alloc_state		= 0,
			.low_state_blk_alloc	= 0,
			.high_state_blk_alloc	= 0,
			.prob			= ~0,
		},
		.pwm = { /* L3C_PROF_WR_MISS_t */
			.low_order		= 11,
			.high_order		= 11,
			.alloc_state		= 1,
			.low_state_blk_alloc	= 1,
			.high_state_blk_alloc	= 1,
			.prob			= 0,
		},
		.prh0 = { /* L3C_PROF_RD_HIT0_t */
			.low_delta		= 0x15,
			.high_delta		= 0x15,
			.new_state		= 0,
			.prob			= 0,
		},
		.pwh0 = { /* L3C_PROF_WR_HIT0_t */
			.low_delta		= 11,
			.high_delta		= 11,
			.new_state		= 1,
			.prob			= 0,
		},
		.prh1 = { /* L3C_PROF_RD_HIT1_t */
			.low_delta		= 0,
			.high_delta		= 0,
			.new_state		= 0,
			.prob			= 0,
		},
		.pwh1 = { /* L3C_PROF_WR_HIT1_t */
			.low_delta		= 11,
			.high_delta		= 11,
			.new_state		= 1,
			.prob			= 0,
		},
	},
};

/*
 * The default profile each L3 profile would get.
 * The current setting would make profile 1 the NVMe suitable profile
 * and the rest of the profiles LRU profile.
 * Note that profile 0 should be configured as LRU as this is the
 * default profile.
 */
static const enum l3_profile_type default_profile[L3_PROFILE_NUM] = {
	[1] = NVME_PROFILE,
};

/* L3 cache base struct. */
struct l3cache {
	struct platform_device *pdev;	/* platform device */
	uint32_t sreg_l3c_tbl;		/* verification table for l3 */
	bool sreg_use_smcs;		/* access regs with smcs if true */

	/*
	 * Reg base addr, will be memmapped if SMCs are not used to
	 * access l3c registers. Otherwise, these will correspond
	 * to physical addresses.
	 */
	void __iomem	*l3_base[L3_PER_MSS];
};

static int secure_writel(struct l3cache *cache, uint32_t value,
			 void __iomem *addr)
{
	struct arm_smccc_res res;
	int status;

	arm_smccc_smc(MLNX_WRITE_REG_32, cache->sreg_l3c_tbl, value,
		(uintptr_t) addr, 0, 0, 0, 0, &res);

	status = res.a0;

	switch (status) {
	/*
	 * Note: PSCI_RET_NOT_SUPPORTED is used here to maintain compatibility
	 * with older kernels that do not have SMCCC_RET_NOT_SUPPORTED
	 */
	case PSCI_RET_NOT_SUPPORTED:
		dev_err(&cache->pdev->dev,
			"%s: required SMC unsupported\n",
			__func__);
		return -1;
	case SMCCC_ACCESS_VIOLATION:
		dev_err(&cache->pdev->dev,
			"%s: could not access register at %p\n",
			__func__,
			addr);
		return -1;
	default:
		return 0;
	}
}

static int l3c_writel(struct l3cache *cache, uint32_t value,
		      void __iomem *addr)
{
	if (cache->sreg_use_smcs)
		return secure_writel(cache, value, addr);
	else {
		writel(value, addr);
		return 0;
	}
}

static ssize_t profiles_show(struct device_driver *drv, char *buf)
{
	int i;
	ssize_t line_size;
	ssize_t len = 0;

	for (i = 0; i < L3_PROFILE_NUM; i++) {
		line_size = sprintf(buf, "%d %s\n", i,
					l3_profiles[default_profile[i]].name);
		buf += line_size;
		len += line_size;
	}
	return len;
}

static DRIVER_ATTR_RO(profiles);

static struct attribute *mlxl3_dev_attrs[] = {
	&driver_attr_profiles.attr,
	NULL
};

static struct attribute_group mlxl3_attr_group = {
	.attrs = mlxl3_dev_attrs,
};

static const struct attribute_group *mlxl3_attr_groups[] = {
	&mlxl3_attr_group,
	NULL,
};

static int l3_register_init(struct l3cache *cache, int reg_idx,
			     enum l3_profile_type type)
{
	int i;
	int status;

	/* Set the registers in both L3C to be the profile's value. */
	for (i = 0; i < L3_PER_MSS; i++) {
		void __iomem *l3_base = cache->l3_base[i];

		status = l3c_writel(cache, l3_profiles[type].prm.word,
			l3_base + L3_REG_OFFSET(PROF_RD_MISS, reg_idx));
		if (status < 0)
			return status;

		status = l3c_writel(cache, l3_profiles[type].pwm.word,
			l3_base + L3_REG_OFFSET(PROF_WR_MISS, reg_idx));
		if (status < 0)
			return status;

		status = l3c_writel(cache, l3_profiles[type].prh0.word,
			l3_base + L3_REG_OFFSET(PROF_RD_HIT0, reg_idx));
		if (status < 0)
			return status;

		status = l3c_writel(cache, l3_profiles[type].pwh0.word,
			l3_base + L3_REG_OFFSET(PROF_WR_HIT0, reg_idx));
		if (status < 0)
			return status;

		status = l3c_writel(cache, l3_profiles[type].prh1.word,
			l3_base + L3_REG_OFFSET(PROF_RD_HIT1, reg_idx));
		if (status < 0)
			return status;

		status = l3c_writel(cache, l3_profiles[type].pwh1.word,
			l3_base + L3_REG_OFFSET(PROF_WR_HIT1, reg_idx));
		if (status < 0)
			return status;
	}

	return 0;
}

/* Probe the L3 cache. */
static int l3cache_probe(struct platform_device *pdev)
{
	struct resource *l3_res[L3_PER_MSS];
	struct l3cache *cache;
	int i;
	struct arm_smccc_res res;
	void __iomem *l3_base_tmp;

	/* Get the resource of all L3 cache regions. */
	for (i = 0; i < L3_PER_MSS; i++) {
		l3_res[i] = platform_get_resource(pdev, IORESOURCE_MEM, i);
		if (!l3_res[i]) {
			dev_err(&pdev->dev, "could not get device resource\n");
			return -EINVAL;
		}
	}

	cache = devm_kzalloc(&pdev->dev, sizeof(struct l3cache), GFP_KERNEL);
	if (!cache) {
		dev_err(&pdev->dev, "could not alloc l3cache data\n");
		return -ENOMEM;
	}

	cache->pdev = pdev;
	platform_set_drvdata(pdev, cache);

	/* Determine whether to use SMCs or not. */
	if (device_property_read_u32(&pdev->dev,
			"sec_reg_block", &cache->sreg_l3c_tbl))
		/* If sec_reg_block not specified, don't use SMCs at all */
		cache->sreg_use_smcs = false;
	else {
		/*
		 * Ensure we have the UUID we expect for the Mellanox service.
		 */
		arm_smccc_smc(MLNX_SIP_SVC_UID, 0, 0, 0, 0, 0, 0, 0, &res);
		if (res.a0 != 0x89c036b4 || res.a1 != 0x11e6e7d7 ||
			res.a2 != 0x1a009787 || res.a3 != 0xc4bf00ca) {
			dev_err(&pdev->dev,
				"Mellanox SMC service not available\n");
			return -EINVAL;
		}

		/*
		 * Check service version to see if we actually do support the
		 * needed SMCs. If we have the calls we need, mark support for
		 * them in the cache struct.
		 */
		arm_smccc_smc(MLNX_SIP_SVC_VERSION, 0, 0, 0, 0, 0, 0, 0, &res);
		if (res.a0 == MLNX_L3C_SVC_REQ_MAJOR &&
			res.a1 >= MLNX_L3C_SVC_MIN_MINOR) {
			cache->sreg_use_smcs = true;
		} else {
			dev_err(&pdev->dev,
				"Required SMCs are not supported.\n");

			return -EINVAL;
		}
	}

	/* Map registers */
	for (i = 0; i < L3_PER_MSS; i++) {
		if (!cache->sreg_use_smcs) {
			l3_base_tmp = devm_ioremap_resource(
					&pdev->dev, l3_res[i]);

			if (IS_ERR(l3_base_tmp)) {
				dev_err(&pdev->dev, "could not map %p\n",
					(void __iomem *) l3_res[i]->start);
				return PTR_ERR(l3_base_tmp);
			}
		} else {
			l3_base_tmp = (void __iomem *) l3_res[i]->start;
		}

		cache->l3_base[i] = l3_base_tmp;
	}

	/* Initialize all the profiles */
	for (i = 0; i < L3_PROFILE_NUM; i++) {
		if (l3_register_init(cache, i, default_profile[i]) < 0) {
			dev_err(&pdev->dev,
				"could not init l3 profile %d\n", i);
			return -EINVAL;
		}
	}

	dev_info(&pdev->dev, "probed\n");
	return 0;
}

static const struct acpi_device_id bf_l3cache_acpi_match[] = {
	{ "MLNXBF07", 0 },
	{},
};
MODULE_DEVICE_TABLE(acpi, bf_l3cache_acpi_match);

static struct platform_driver l3cache_driver = {
	.probe = l3cache_probe,
	.driver = {
		.name = DRIVER_NAME,
		.groups = mlxl3_attr_groups,
		.acpi_match_table = ACPI_PTR(bf_l3cache_acpi_match),
	},
};

static int __init l3cache_init(void)
{
	int ret;

	ret = platform_driver_register(&l3cache_driver);
	if (ret)
		pr_err("Failed to register l3cache driver.\n");

	return ret;
}

static void __exit l3cache_exit(void)
{
	platform_driver_unregister(&l3cache_driver);
}

module_init(l3cache_init);
module_exit(l3cache_exit);

MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_AUTHOR("Mellanox Technologies, Ltd");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);
